"""
Pull resources from a cloud provider account.
"""

from threading import Lock, Thread, current_thread
from queue import Queue
from typing import Literal, Union

from ..base.call_api import call_api
from ..base.base_list import BaseList
from ..auth.token import BasicAuth
from ..exceptions.Exceptions import *
from .data_classes.AWSResources import *
from .data_classes.resource_mappings import *


termination_flag = False


def fetch_page(
    auth: BasicAuth,
    provider: str,
    resourceType: str,
    pageNo: int,
    results: BaseList,
    lock: Lock,
    page_count: Union[int, "all"],
    **kwargs,
):
    """
    Worker function to fetch a single page of resources.
    """
    global termination_flag

    kwargs["placeholder"] = resourceType
    kwargs["cloudprovider"] = provider

    match kwargs["cloudprovider"].upper():
        case "AWS":
            kwargs["cloudprovider"] = "AWS"
        case "AZURE":
            kwargs["cloudprovider"] = "Azure"
        case _:
            raise ValueError("Invalid provider. Must be 'aws' or 'azure'.")

    kwargs["pageNo"] = pageNo

    response = call_api(
        auth=auth,
        module="cloudview",
        endpoint="get_inventory",
        params=kwargs,
    )

    if response.status_code not in [200, 400, 404]:
        if not response.text:
            raise QualysAPIError(
                "An error occurred while retrieving the inventory. Most likely, a parameter you set is incorrect."
            )
        else:
            raise QualysAPIError(response.json())

    # Check for empty response or invalid page count
    if (
        not response.text
        or response.status_code in [400, 404]
        or (page_count != "all" and pageNo >= 199)
    ):
        with lock:
            termination_flag = True
        return

    j = response.json()

    if j.get("empty", True):
        with lock:
            termination_flag = True
        return

    with lock:
        for i in j["content"]:
            resource_class = resource_map.get(resourceType, None)
            if resource_class:
                if "type" in i.keys():
                    i["_type"] = i.pop("type")
                i = resource_class(**i)
            else:
                raise ValueError(
                    f"Invalid resource type {resourceType} for provider {provider}. Valid resource types are:\n{VALID_RESOURCETYPES[provider]}"
                )
            results.append(i)

        if (pageNo + 1) % 20 == 0:
            print(
                f"({current_thread().name}) Page {pageNo+1} of {provider}-{resourceType} retrieved successfully."
            )

        if "INTERNAL_SERVER_ERROR" in response.text and response.status_code == 502:
            # We have reached the end of the pages
            with lock:
                termination_flag = True


def worker(
    auth,
    provider,
    resourceType,
    results,
    queue,
    lock,
    page_count,
    **kwargs,
):
    global termination_flag

    while True:
        if termination_flag:
            break

        pageNo = queue.get()
        if pageNo is None:
            break

        fetch_page(
            auth,
            provider,
            resourceType,
            pageNo,
            results,
            lock,
            page_count,
            **kwargs,
        )
        queue.task_done()


def check_termination_or_empty(queue):
    return queue.empty() or termination_flag


def get_inventory(
    auth: BasicAuth,
    provider: Literal["aws", "azure"],
    resourceType: str,
    page_count: Union[int, "all"] = "all",
    thread_count: int = 5,
    **kwargs,
) -> BaseList:
    """
    Get resources from a cloud provider account.

    Args:
        auth (BasicAuth): The authentication object.
        provider (Literal["aws", "azure"]): The cloud provider to get resources from.
        resourceType (str): The type of resource to get.
        page_count (Union[int, "all"]): The number of pages to return. MAX VALUE IS 200. If 'all', return all pages. Default is 'all'.
        thread_count (int): The number of threads to use for fetching data.

     ## Kwargs:

         sort (Literal['lastSyncedOn:asc', 'lastSyncedOn:desc']): Sort the resources by lastSyncedOn in ascending or descending order.
         updated (str): Filter resources by the last updated date. Format is Qualys QQL, like [2024-01-01 ... 2024-12-31], [2024-01-01 ... now-1m] (month), 2024-01-01, etc.
         filter (str): Filter resources by providing a Qualys QQL query.


    Returns:
        BaseList: The response from the API as a BaseList of Resource objects.
    """

    global termination_flag
    termination_flag = False

    # Validate provider
    provider = provider.lower()
    resourceType = resourceType.upper()

    kwargs["pageSize"] = 50

    if provider not in ["aws", "azure"]:
        raise ValueError("Invalid provider. Must be 'aws' or 'azure'.")

    # Handle common names
    resourceType = resourceType.replace(" ", "_")
    for key, value in COMMON_NAMES[provider].items():
        if resourceType in value:
            resourceType = key
            break

    if resourceType not in VALID_RESOURCETYPES[provider]:
        raise ValueError(
            f"Invalid resource type for provider {provider}. Valid resource types are: {VALID_RESOURCETYPES[provider]}"
        )

    if page_count != "all" and (not isinstance(page_count, int)):
        raise ValueError("page_count must be an integer <= 200 or 'all'.")

    # If user has set page_count to a number >= 200, set it to 199
    if page_count != "all" and page_count >= 200:
        page_count = 199
    elif page_count != "all" and page_count < 1:
        raise ValueError("page_count must be an integer >= 1.")

    if not isinstance(thread_count, int) or thread_count < 1:
        raise ValueError("thread_count must be an integer >= 1.")

    results = BaseList()
    page_queue = Queue()
    lock = Lock()

    # Pre-fill the queue with page numbers
    for i in range(200) if page_count == "all" else range(page_count):
        page_queue.put(i)

    # Start worker threads
    threads = []
    for _ in range(thread_count):
        t = Thread(
            target=worker,
            args=(auth, provider, resourceType, results, page_queue, lock, page_count),
            kwargs=kwargs,
        )
        threads.append(t)

    for t in threads:
        t.start()

    # Wait for the queue to be empty, or the termination flag to be set
    while not check_termination_or_empty(page_queue):
        pass

    # Stop the worker threads
    for _ in range(thread_count):
        page_queue.put(None)
    for t in threads:
        t.join()

    # print(f"{str(len(results))} {provider} {resourceType} records retrieved.")
    return results



"""
Contains the WebAppAuthServerRecord class for the WAS
"""

from dataclasses import dataclass, asdict
from typing import Union


@dataclass
class WebAppAuthServerRecord:
    """
    Represents a single server authentication record
    in Qualys WAS
    """

    id: Union[str, int] = None
    type: str = None
    domain: str = None
    username: str = None
    password: str = None

    def __post_init__(self):
        if not isinstance(self.id, (str, int)):
            raise ValueError("id must be a string or integer")
        setattr(self, "id", int(self.id))

    def __str__(self):
        return f"{self.type} : {self.username}"

    def to_dict(self):
        return asdict(self)

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)

    def redact_password(self):
        """
        Obfuscates the password
        """
        if self.password:
            setattr(self, "password", "REDACTED")



"""
xml_parser.py - contains the xml_parser function that parses an XML string into a dictionary.
"""

from lxml.etree import _Comment
from defusedxml.lxml import fromstring


def xml_parser(xml_string, attr_prefix="@", cdata_key="#text"):
    """
    Turn an xml string into a dictionary.

     Params:
         xml_string (str): The xml string to parse.
         attr_prefix (str): The prefix to add to attributes.
         cdata_key (str): The key to use for cdata.

     Returns:
         dict: The parsed xml as a dictionary.
    """
    # check if the user passed in a string or bytes. if string, encode it to utf-8
    if isinstance(xml_string, str):
        xml_string = xml_string.encode("utf-8")

    def parse_element(element):
        parsed_dict = {}
        # Parse attributes
        for key, value in element.attrib.items():
            parsed_dict[attr_prefix + key] = value
        # Parse child elements
        for child in element:
            if isinstance(child, _Comment):
                continue  # Skip comments
            child_dict = parse_element(child)
            if child.tag in parsed_dict:
                if not isinstance(parsed_dict[child.tag], list):
                    parsed_dict[child.tag] = [parsed_dict[child.tag]]
                parsed_dict[child.tag].append(child_dict)
            else:
                parsed_dict[child.tag] = child_dict
        # Parse text content
        text = (element.text or "").strip()
        if text:
            if parsed_dict:
                parsed_dict[cdata_key] = text
            else:
                parsed_dict = text
        return parsed_dict

    root = fromstring(xml_string)
    return {root.tag: parse_element(root)}



"""
scanner_appliance.py - Contains the ScannerAppliance data class.
"""

from dataclasses import dataclass, field, asdict
from uuid import UUID as uuid
from typing import Literal
from datetime import datetime
from ipaddress import IPv4Address

from ...base.base_list import BaseList
from .asset_group import AssetGroup
from .tag import Tag


def build_dataclass_baselist(
    data: list[dict], class_type: Literal["ag", "tag"]
) -> BaseList:
    """
    Helps to normalize and build a BaseList of dataclasses from a list of dictionaries
    for the various fields in the ScannerAppliance dataclass.

    Parameters:
        data (list[dict]): The list of dictionaries to convert to a BaseList of dataclasses.
        class_type (Literal["ag", "tag"]): The type of dataclass to build. Either "ag" for AssetGroup or "tag" for Tag.

    Returns:
        BaseList[Union[AssetGroup, Tag]]: A BaseList of dataclasses.
    """

    key_mapping = {
        "ag": "ASSET_GROUP",
        "tag": "ASSET_TAG",
    }

    result = BaseList()

    # Qualys can return an immediate dict instead of a list of dicts
    # if the data is only one item. Convert it to a list of one dict
    # for normalization.

    if isinstance(data[key_mapping[class_type]], dict):
        data[key_mapping[class_type]] = [data[key_mapping[class_type]]]

    for item in data[key_mapping[class_type]]:
        match class_type:
            case "ag":
                # Convert the NAME key to TITLE
                item["TITLE"] = item.pop("NAME")
                # And build the AssetGroup object
                item = AssetGroup(**item)
            case "tag":
                # Convert the UUID key to ID
                item["TAG_ID"] = uuid(item.pop("UUID"))
                # And build the Tag object
                item = Tag(**item)
            # Other cases coming soon
        result.append(item)

    return result


@dataclass
class SecurityGroup:
    """
    Dataclass to store a scanner appliance's cloud security group information.
    """

    SECURITY_GROUP_IDS: str = field(
        metadata={"description": "The security group ID."},
        default=None,
    )

    SECURITY_GROUP_NAMES: str = field(
        metadata={"description": "The security group name."},
        default=None,
    )

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


@dataclass
class ScannerCloudInfo:
    """
    Dataclass to store a scanner appliance's cloud provider
    and instance information.
    """

    PLATFORM_PROVIDER: str = field(
        metadata={"description": "The cloud platform provider."},
        default=None,
    )

    EC2_INFO: dict = field(
        metadata={"description": "EC2 information. Gets deleted out."},
        default=None,
    )

    INSTANCE_ID: str = field(
        metadata={"description": "The cloud instance ID."},
        default=None,
    )

    INSTANCE_TYPE: str = field(
        metadata={"description": "The cloud instance type."},
        default=None,
    )

    AMI_ID: str = field(
        metadata={"description": "The cloud AMI ID."},
        default=None,
    )

    ACCOUNT_ID: int = field(
        metadata={"description": "The cloud account ID."},
        default=None,
    )

    INSTANCE_REGION: str = field(
        metadata={"description": "The cloud instance region."},
        default=None,
    )

    INSTANCE_AVAILABILITY_ZONE: str = field(
        metadata={"description": "The cloud instance availability zone."},
        default=None,
    )

    INSTANCE_ZONE_TYPE: str = field(
        metadata={"description": "The cloud instance zone type."},
        default=None,
    )

    INSTANCE_VPC_ID: str = field(
        metadata={"description": "The cloud instance VPC ID."},
        default=None,
    )

    INSTANCE_SUBNET_ID: str = field(
        metadata={"description": "The cloud instance subnet ID."},
        default=None,
    )

    INSTANCE_ADDRESS_PRIVATE: IPv4Address = field(
        metadata={"description": "The cloud instance private address."},
        default=None,
    )

    INSTANCE_ADDRESS_PUBLIC: IPv4Address = field(
        metadata={"description": "The cloud instance public address."},
        default=None,
    )

    HOSTNAME_PRIVATE: str = field(
        metadata={"description": "The cloud instance private hostname."},
        default=None,
    )

    SECURITY_GROUPS: BaseList[SecurityGroup] = field(
        metadata={"description": "The cloud instance security groups."},
        default=None,
    )

    API_PROXY_SETTINGS: Literal["Disabled", "Enabled"] = field(
        metadata={"description": "The cloud instance API proxy settings."},
        default=None,
    )

    def __post_init__(self):
        """
        Post init function to convert certain fields to int/float.
        """
        INT_FIELDS = ["ACCOUNT_ID"]
        IPV4_FIELDS = ["INSTANCE_ADDRESS_PRIVATE", "INSTANCE_ADDRESS_PUBLIC"]

        if self.EC2_INFO:
            # Set the parsed out attributes
            self.INSTANCE_ID = self.EC2_INFO.get("INSTANCE_ID")
            self.INSTANCE_TYPE = self.EC2_INFO.get("INSTANCE_TYPE")
            self.AMI_ID = self.EC2_INFO.get("AMI_ID")
            self.ACCOUNT_ID = self.EC2_INFO.get("ACCOUNT_ID")
            self.INSTANCE_REGION = self.EC2_INFO.get("INSTANCE_REGION")
            self.INSTANVE_AVAILABILITY_ZONE = self.EC2_INFO.get(
                "INSTANVE_AVAILABILITY_ZONE"
            )
            self.INSTANCE_ZONE_TYPE = self.EC2_INFO.get("INSTANCE_ZONE_TYPE")
            self.INSTANCE_VPC_ID = self.EC2_INFO.get("INSTANCE_VPC_ID")
            self.INSTANCE_SUBNET_ID = self.EC2_INFO.get("INSTANCE_SUBNET_ID")
            self.INSTANCE_ADDRESS_PRIVATE = self.EC2_INFO.get(
                "INSTANCE_ADDRESS_PRIVATE"
            )
            self.INSTANCE_ADDRESS_PUBLIC = self.EC2_INFO.get("INSTANCE_ADDRESS_PUBLIC")
            self.HOSTNAME_PRIVATE = self.EC2_INFO.get("HOSTNAME_PRIVATE")
            if self.EC2_INFO.get("SECURITY_GROUPS"):
                self.SECURITY_GROUPS = SecurityGroup.from_dict(
                    self.EC2_INFO.get("SECURITY_GROUPS")
                )
            self.API_PROXY_SETTINGS = self.EC2_INFO.get("API_PROXY_SETTINGS")["SETTING"]

        for field in INT_FIELDS:
            if getattr(self, field):
                setattr(self, field, int(getattr(self, field)))

        for field in IPV4_FIELDS:
            if getattr(self, field):
                setattr(self, field, IPv4Address(getattr(self, field)))

        del self.EC2_INFO

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)

    def to_dict(self):
        return asdict(self)

    def keys(self):
        return asdict(self).keys()

    def values(self):
        return asdict(self).values()

    def items(self):
        return asdict(self).items()

    def __getitem__(self, key):
        return asdict(self)[key]

    def __setitem__(self, key, value):
        asdict(self)[key] = value
        return asdict(self)

    def __str__(self):
        return self.INSTANCE_ID

    def __repr__(self) -> str:
        # Only return non-empty values
        return (
            "ScannerCloudInfo("
            + ", ".join(f"{k}={v}" for k, v in self.valid_values().items())
            + ")"
        )

    def valid_values(self):
        return {k: v for k, v in asdict(self).items() if v}


@dataclass
class ProxySettings:
    """
    Proxy settings for a scanner appliance.
    """

    SETTING: str = field(
        metadata={"description": "The setting."},
        default=None,
    )

    PROXY: dict = field(
        metadata={"description": "The proxy settings."},
        default=None,
    )

    def __str__(self) -> str:
        return self.SETTING

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


@dataclass
class InterfaceSettings:
    """
    Dataclass to store a scanner appliance's interface settings.
    """

    INTERFACE: str = field(
        metadata={"description": "The interface."},
        default=None,
    )

    IP_ADDRESS: IPv4Address = field(
        metadata={"description": "The IP address."},
        default=None,
    )

    NETMASK: str = field(
        metadata={"description": "The netmask."},
        default=None,
    )

    GATEWAY: IPv4Address = field(
        metadata={"description": "The gateway."},
        default=None,
    )

    LEASE: str = field(
        metadata={"description": "The lease."},
        default=None,
    )

    SPEED: int = field(
        metadata={"description": "The speed."},
        default=None,
    )

    DUPLEX: str = field(
        metadata={"description": "The duplex."},
        default=None,
    )

    DNS: dict = field(
        metadata={"description": "The DNS settings."},
        default=None,
    )

    DNS_PRIMARY: IPv4Address = field(
        metadata={"description": "The primary DNS."},
        default=None,
    )

    DNS_SECONDARY: IPv4Address = field(
        metadata={"description": "The secondary DNS."},
        default=None,
    )

    SETTING: str = field(
        metadata={"description": "The setting."},
        default=None,
    )

    DOMAIN: str = field(
        metadata={"description": "The domain."},
        default=None,
    )

    def __post_init__(self):
        """
        Post init function to convert certain fields to int/float.
        """
        INT_FIELDS = ["SPEED"]
        IPV4_FIELDS = ["IP_ADDRESS", "GATEWAY"]

        for field in INT_FIELDS:
            if getattr(self, field):
                setattr(self, field, int(getattr(self, field)))

        for field in IPV4_FIELDS:
            if getattr(self, field):
                setattr(self, field, IPv4Address(getattr(self, field)))

        if self.DNS:
            self.DNS_PRIMARY = self.DNS.get("PRIMARY")
            self.DNS_SECONDARY = self.DNS.get("SECONDARY")
            self.DOMAIN = self.DNS.get("DOMAIN")
            del self.DNS

    def __str__(self):
        # Comma-separated string of non-empty values in k:v format
        return ", ".join(f"{k}:{v}" for k, v in self.valid_values().items() if v)

    def __repr__(self) -> str:
        # Only return non-empty values
        return (
            "InterfaceSettings("
            + ", ".join(f"{k}={v}" for k, v in self.valid_values().items())
            + ")"
        )

    def valid_values(self):
        return {k: v for k, v in asdict(self).items() if v}

    def keys(self):
        return asdict(self).keys()

    def values(self):
        return asdict(self).values()

    def items(self):
        return asdict(self).items()

    def __getitem__(self, key):
        return asdict(self)[key]

    def __setitem__(self, key, value):
        asdict(self)[key] = value
        return asdict(self)

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)

    def __dict__(self):
        return asdict(self)


@dataclass(order=True)
class ScannerAppliance:
    """
    Scanner Appliances in VMDR.
    """

    ID: int = field(
        metadata={"description": "Scanner Appliance ID."},
    )

    NAME: str = field(
        metadata={"description": "Scanner Appliance friendly name."},
        default=None,
    )

    SOFTWARE_VERSION: float = field(
        metadata={"description": "Software version of the scanner appliance."},
        default=None,
    )

    RUNNING_SCAN_COUNT: int = field(
        metadata={
            "description": "Number of scans currently running on the scanner appliance."
        },
        default=0,
    )

    STATUS: str = field(
        metadata={"description": "Status of the scanner appliance."},
        default=None,
    )

    UUID: uuid = field(
        metadata={"description": "UUID of the scanner appliance."},
        default=None,
    )

    RUNNING_SLICES_COUNT: int = field(
        metadata={
            "description": "Number of slices currently running on the scanner appliance."
        },
        default=0,
    )

    MODEL_NUMBER: str = field(
        metadata={"description": "Model number of the scanner appliance."},
        default=None,
    )

    TYPE: str = field(
        metadata={"description": "Type of scanner appliance."},
        default=None,
    )

    SERIAL_NUMBER: str = field(
        metadata={"description": "Serial number of the scanner appliance."},
        default=None,
    )

    ACTIVATION_CODE: int = field(
        metadata={"description": "Activation code of the scanner appliance."},
        default=None,
    )

    INTERFACE_SETTINGS: InterfaceSettings = field(
        metadata={"description": "Interface settings of the scanner appliance."},
        default=None,
    )

    PROXY_SETTINGS: ProxySettings = field(
        metadata={"description": "Proxy settings of the scanner appliance."},
        default=None,
    )

    IS_CLOUD_DEPLOYED: bool = field(
        metadata={"description": "Whether the scanner appliance is cloud deployed."},
        default=False,
    )

    CLOUD_INFO: ScannerCloudInfo = field(
        metadata={"description": "Cloud information of the scanner appliance."},
        default=None,
    )

    VLANS: dict = field(
        metadata={"description": "VLANs of the scanner appliance."},
        default=None,
    )

    STATIC_ROUTES: dict = field(
        metadata={"description": "Static routes of the scanner appliance."},
        default=None,
    )

    ML_LATEST: str = field(
        metadata={
            "description": "Latest machine learning model of the scanner appliance."
        },
        default=None,
    )

    ML_VERSION: dict = field(
        metadata={"description": "Machine learning version of the scanner appliance."},
        default=None,
    )

    VULNSIGS_LATEST: str = field(
        metadata={
            "description": "Latest vulnerability signature of the scanner appliance."
        },
        default=None,
    )

    VULNSIGS_VERSION: dict = field(
        metadata={
            "description": "Vulnerability signature version of the scanner appliance."
        },
        default=None,
    )

    ASSET_GROUP_COUNT: int = field(
        metadata={
            "description": "Number of asset groups associated with the scanner appliance."
        },
        default=0,
    )

    ASSET_GROUP_LIST: BaseList[AssetGroup] = field(
        metadata={
            "description": "List of asset groups associated with the scanner appliance."
        },
        default=None,
    )

    ASSET_TAGS_LIST: BaseList[Tag] = field(
        metadata={
            "description": "List of asset tags associated with the scanner appliance."
        },
        default=None,
    )

    LAST_UPDATED_DATE: datetime = field(
        metadata={"description": "Date the scanner appliance was last updated."},
        default=None,
    )

    POLLING_INTERVAL: int = field(
        metadata={"description": "Polling interval of the scanner appliance."},
        default=None,
    )

    USER_LOGIN: str = field(
        metadata={"description": "Owner account of the scanner appliance."},
        default=None,
    )

    HEARTBEATS_MISSED: int = field(
        metadata={
            "description": "Number of heartbeats missed by the scanner appliance."
        },
        default=None,
    )

    SS_CONNECTION: str = field(
        metadata={
            "description": "Status of the scanner appliance's connection to the Security Server."
        },
        default=None,
    )

    SS_LAST_CONNECTED: datetime = field(
        metadata={
            "description": "Date the scanner appliance last connected to the Security Server."
        },
        default=None,
    )

    USER_LIST: dict = field(
        metadata={
            "description": "List of users associated with the scanner appliance."
        },
        default=None,
    )

    UPDATED: bool = field(
        metadata={"description": "Whether the scanner appliance was updated."},
        default=None,
    )

    COMMENTS: str = field(
        metadata={"description": "Comments on the scanner appliance."},
        default=None,
    )

    MAX_CAPACITY_UNITS: int = field(
        metadata={"description": "Maximum capacity units of the scanner appliance."},
        default=None,
    )

    # BELOW ATTRIBUTES ARE PULLED UP A LEVEL FROM AN ABOVE ATTRIBUTE DICITONARY

    def __post_init__(self):
        """
        Post init function to convert certain fields to int/float.
        """
        INT_FIELDS = [
            "ID",
            "RUNNING_SCAN_COUNT",
            "RUNNING_SLICES_COUNT",
            "ACTIVATION_CODE",
            "ASSET_GROUP_COUNT",
            "POLLING_INTERVAL",
            "HEARTBEATS_MISSED",
            "MAX_CAPACITY_UNITS",
        ]
        DT_FIELDS = ["LAST_UPDATED_DATE", "SS_LAST_CONNECTED"]
        FLOAT_FIELDS = ["SOFTWARE_VERSION"]
        CUSTOM_DATACLASSES = [("ASSET_GROUP_LIST", "ag"), ("ASSET_TAGS_LIST", "tag")]

        for field in INT_FIELDS:
            if getattr(self, field):
                # Special case for polling interval. Split by space and take the first value.
                if field == "POLLING_INTERVAL":
                    setattr(self, field, int(getattr(self, field).split()[0]))
                else:
                    setattr(self, field, int(getattr(self, field)))

        for field in FLOAT_FIELDS:
            if getattr(self, field):
                setattr(self, field, float(getattr(self, field)))

        for field in DT_FIELDS:
            if getattr(self, field):
                setattr(
                    self,
                    field,
                    datetime.strptime(getattr(self, field), "%Y-%m-%dT%H:%M:%S%z"),
                )

        for field in CUSTOM_DATACLASSES:
            if getattr(self, field[0]):
                setattr(
                    self,
                    field[0],
                    build_dataclass_baselist(getattr(self, field[0]), field[1]),
                )

        if self.CLOUD_INFO:
            setattr(self, "CLOUD_INFO", ScannerCloudInfo.from_dict(self.CLOUD_INFO))

        # convert UUID:
        if self.UUID:
            setattr(self, "UUID", uuid(self.UUID))

        if self.UPDATED:
            setattr(self, "UPDATED", self.UPDATED == "Yes")

        # Convert IS_CLOUD_DEPLOYED to bool:
        if self.IS_CLOUD_DEPLOYED:
            setattr(self, "IS_CLOUD_DEPLOYED", self.IS_CLOUD_DEPLOYED == "1")

        if self.INTERFACE_SETTINGS:
            if isinstance(self.INTERFACE_SETTINGS, dict):
                setattr(
                    self,
                    "INTERFACE_SETTINGS",
                    InterfaceSettings.from_dict(self.INTERFACE_SETTINGS),
                )
            else:
                setattr(
                    self,
                    "INTERFACE_SETTINGS",
                    BaseList(
                        [
                            InterfaceSettings.from_dict(item)
                            for item in self.INTERFACE_SETTINGS
                        ]
                    ),
                )

        if self.PROXY_SETTINGS:
            if isinstance(self.PROXY_SETTINGS, dict):
                setattr(
                    self, "PROXY_SETTINGS", ProxySettings.from_dict(self.PROXY_SETTINGS)
                )
            else:
                setattr(
                    self,
                    "PROXY_SETTINGS",
                    BaseList(
                        [ProxySettings.from_dict(item) for item in self.PROXY_SETTINGS]
                    ),
                )

    def __str__(self):
        return self.NAME

    def __int__(self):
        return self.ID

    def __dict__(self):
        return asdict(self)

    def keys(self):
        return asdict(self).keys()

    def values(self):
        return asdict(self).values()

    def items(self):
        return asdict(self).items()

    def __getitem__(self, key):
        return asdict(self)[key]

    def __setitem__(self, key, value):
        asdict(self)[key] = value
        return asdict(self)

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)

    def to_dict(self):
        return asdict(self)

    def valid_values(self):
        return {k: v for k, v in asdict(self).items() if v}



"""
convert_bools_and_nones.py - Contains functions to convert bools and Nones to 1/0 and 'None' respectively. 

Used with the various kwargs an API endpoint accepts to get them into a requests friendly format.
"""


def convert_bools_and_nones(kwargs: dict) -> dict:
    """
    Converts bools to 1/0 and Nones to 'None' in a dictionary.

    Params:
        kwargs (dict): Dictionary of keyword arguments.

    Returns:
        dict: Dictionary with bools converted to 1/0 and Nones converted to 'None'.
    """
    # If any kwarg is a bool, convert it to 1 or 0
    for key in kwargs:
        if isinstance(kwargs[key], bool):
            kwargs[key] = 1 if kwargs[key] else 0

    # If any kwarg is None, set it to 'None'
    for key in kwargs:
        if kwargs[key] is None:
            kwargs[key] = "None"

        # if host_metadata is specified and set to a non-"all" or non-NoneType, lower() it:
        if key == "host_metadata" and kwargs[key] not in ["all", None]:
            kwargs[key] = kwargs[key].lower()

    return kwargs



"""
CLI script to quickly perform Web Application Scanning
(WAS) operations using qualysdk.
"""

from argparse import ArgumentParser

from qualysdk import BasicAuth, write_csv
from qualysdk.was import *
from qualysdk.sql.base import prepare_dataclass
from pandas import DataFrame


def cli_findings(auth: BasicAuth, args) -> None:
    kwargs = dict(args.kwarg) if args.kwarg else {}
    if "page_count" in kwargs:
        kwargs["page_count"] = int(kwargs["page_count"])
    for kwarg in kwargs:
        if str(kwargs[kwarg]).lower() == "true":
            kwargs[kwarg] = True
        elif str(kwargs[kwarg]).lower() == "false":
            kwargs[kwarg] = False

    findings = get_findings(auth, **kwargs)
    write_csv(findings, args.output)


def main():
    parser = ArgumentParser(
        description="CLI script to quickly perform Web Application Scanning (WAS) operations using qualysdk"
    )
    parser.add_argument(
        "-u", "--username", required=True, help="Qualys username", type=str
    )
    parser.add_argument(
        "-p", "--password", required=True, help="Qualys password", type=str
    )
    parser.add_argument(
        "-P",
        "--platform",
        help="Qualys platform",
        default="qg3",
        choices=["qg1", "qg2", "qg3", "qg4"],
    )

    # subparser for action:
    subparsers = parser.add_subparsers(dest="action", help="Action to perform")

    # list findings:
    list_findings_parser = subparsers.add_parser(
        "get_findings", help="Get a list of WAS findings."
    )
    list_findings_parser.add_argument(
        "-o",
        "--output",
        help="Output CSV file to write results to",
        type=str,
        default="was_findings.csv",
    )
    list_findings_parser.add_argument(
        "--kwarg",
        help="Specify a keyword argument to pass to the get_findings method. Can be used multiple times",
        action="append",
        nargs=2,
        metavar=("key", "value"),
    )

    args = parser.parse_args()

    # create BasicAuth object
    auth = BasicAuth(args.username, args.password, platform=args.platform)

    # perform action
    if args.action == "get_findings":
        cli_findings(auth, args)
    else:
        parser.print_help()
        exit(1)


if __name__ == "__main__":
    main()



"""
Enable or disable printing of stdout.

enable_stdout -> bool: Enable printing.
disable_stdout -> bool: Disable printing.
"""

import sys, os

disable_stdout = lambda: setattr(sys, "stdout", open(os.devnull, "w"))
enable_stdout = lambda: setattr(sys, "stdout", sys.__stdout__)



"""
prepare_criteria.py - backend function to prepare XML string (Criteria tag) for Qualys Cloud Agent API calls
"""

import datetime


def prepare_criteria(**kwargs):
    """
    Prepare the XML string for the filters to apply to the API call.

    Args:
        **kwargs: The filters to apply to the API call.

    ## Kwargs:

        asset_id (list[str]): A list of asset IDs to filter.
        name (list[str]): A list of asset names to filter.
        created (Union[str, datetime.datetime]): The creation date of the asset.
        updated (Union[str, datetime.datetime]): The updated date of the asset.
        created_operator (Literal['GREATER', 'LESSER']): The operator to apply to the created date.
        updated_operator (Literal['GREATER', 'LESSER']): The operator to apply to the updated date.
        tagName (list[str]): A list of tag names to filter assets under.
        agentUuid (list[str]): A list of agent UUIDs to filter.
        pagination_id (int): The pagination ID to use for the API call.

    Returns:
        str: The XML string to pass to the API call.
    """
    # Prepare the base XML format
    xml = '<?xml version="1.0" encoding="UTF-8" ?><ServiceRequest><filters>{REPLACE}</filters></ServiceRequest>'

    DATE_FIELDS = [
        "created",
        "updated",
        "lastVulnScan",
        "lastComplianceScan",
        "informationGatheredUpdated",
        "lastComplianceScan",
        "lastSystemBoot",
        "lastLoggedOnUser",
        "vulnsUpdated",
        "lastCheckedIn",
        "update",
    ]

    # Add the filters to the XML

    # First, check if created and updated are datetime objects
    for field in DATE_FIELDS:
        if field in kwargs and isinstance(kwargs[field], datetime.datetime):
            kwargs[field] = kwargs[field].strftime("%Y-%m-%dT%H:%M:%S.%fZ")

    # For each kwarg filter, add it to the list value under the Criteria tag,
    # with field=kwarg and value=kwarg value
    # as well as a operator attribute with value="EQUALS" (unless
    # the kwarg is created or updated, in which case the created/updated_operator parameter):
    user_xml = ""

    # Map the kwargs to the tag name Qualys expects
    tag_lookup = {
        "asset_id": "id",
        "pagination_id": "id",
    }

    for kwarg, value in (
        (kwarg, value)
        for kwarg, value in kwargs.items()
        if "operator" not in kwarg and "pagination_id" not in kwarg
    ):
        if not isinstance(value, list) and kwarg not in DATE_FIELDS:
            criteria = f'<Criteria field="{tag_lookup.get(kwarg, kwarg)}" operator="EQUALS">{str(value)}</Criteria>'

        elif kwarg in DATE_FIELDS:
            operator = kwargs.get(
                f"{kwarg}_operator", "EQUALS"
            )  # Fallback to EQUALS if not provided for date fields
            criteria = (
                f'<Criteria field="{kwarg}" operator="{operator}">{value}</Criteria>'
            )

        elif isinstance(value, list):
            # Otherwise, if the value is a list, use the IN operator
            # and comma-sep strings for the values:
            criteria = f'<Criteria field="{tag_lookup.get(kwarg, kwarg)}" operator="IN">{",".join([str(v) for v in value])}</Criteria>'

        else:
            raise ValueError(f"Invalid value for {kwarg}: {value}")

        user_xml += criteria

    if kwargs.get("pagination_id"):
        user_xml += f'<Criteria field="id" operator="GREATER">{kwargs["pagination_id"]}</Criteria>'

    # Replace the {REPLACE} placeholder in the base XML with the user XML:
    return xml.format(REPLACE=user_xml)



"""
calls.py - contains the user-facing functions for most cloud agent API calls.
"""

from typing import Union, Literal

from .data_classes.Agent import CloudAgent
from .prepare_criteria import prepare_criteria
from ..base.call_api import call_api
from ..base.xml_parser import xml_parser
from ..auth.basic import BasicAuth
from ..base.base_list import BaseList

TRANSLATION = {
    "inv": "Inventory_Scan",
    "vuln": "Vulnerability_Scan",
    "pc": "PolicyCompliance_Scan",
    "udc": "UDC_Scan",
    "sca": "SCA_Scan",
    "swca": "SWCA_Scan",
}


def list_agents(
    auth: BasicAuth, page_count: Union[int, "all"] = "all", **kwargs
) -> BaseList[CloudAgent]:
    """
    Get a list of Cloud Agents in your subscription, according to the filters provided.

    Args:
        auth (BasicAuth): The authentication object containing the user's credentials.
        page_count (Union[int, 'all']): The number of pages to retrieve. Defaults to 'all'.
        **kwargs: The filters to apply to the list.

    ## Kwargs:

        qwebHostId (int): The QWEB Host ID to filter the agents by.
        lastVulnScan (str): The date of the last vulnerability scan. Formatted like YYYY-MM-DD[THH:MM:SSZ].
        lastComplianceScan (str): The date of the last compliance scan. Formatted like YYYY-MM-DD[THH:MM:SSZ].
        informationGatheredUpdated (str): The date the information was last updated. Formatted like YYYY-MM-DD[THH:MM:SSZ].
        os (str): The operating system of the agent.
        dnsHostName (str): The DNS hostname of the agent.
        address (str): The IP address of the agent.
        vulnsUpdated (str): The date the vulnerabilities were last updated. Formatted like YYYY-MM-DD[THH:MM:SSZ].
        id (int): The **Asset ID** of the agent.
        name (str): The name of the agent.
        created (str): The creation date of the agent. Formatted like YYYY-MM-DD[THH:MM:SSZ].
        type (str): The type of agent.
        netbiosName (str): The NetBIOS name of the agent.
        netbiosNetworkID (str): The NetBIOS network ID of the agent.
        networkGuid (str): The network GUID of the agent.
        trackingMethod (Literal['NONE', 'IP', 'DNSNAME', 'NETBIOS', 'INSTANCE_ID', 'QAGENT']): The tracking method of the agent.
        port (int): The port of the agent.
        installedSoftware (str): Search installed software on the agent.
        tagName (str): The tag name to filter the agents by.
        tagId (int): The tag ID to filter the agents by.
        update: (str): The date the agent was last updated. Formatted like YYYY-MM-DD[THH:MM:SSZ].
        activationKey (str): The activation key of the agent.
        agentConfigurationName (str): The agent configuration name.
        agentConfigurationId (float): The agent configuration ID.
        agentVersion (str): The agent VB version.
        lastCheckedIn (str): The date the agent last checked in. Formatted like YYYY-MM-DD[THH:MM:SSZ].

        DATE OPERATOR KWARGS:
        created_operator (Literal['GREATER', 'LESSER']): The operator to apply to the created date.
        update_operator (Literal['GREATER', 'LESSER']): The operator to apply to the updated date.
        lastVulnScan_operator (Literal['GREATER', 'LESSER']): The operator to apply to the last vulnerability scan date.
        lastComplianceScan_operator (Literal['GREATER', 'LESSER']): The operator to apply to the last compliance scan date.
        informationGatheredUpdated_operator (Literal['GREATER', 'LESSER']): The operator to apply to the information gathered updated date.
        vulnsUpdated_operator (Literal['GREATER', 'LESSER']): The operator to apply to the vulnerabilities updated date.
        lastCheckedIn_operator (Literal['GREATER', 'LESSER']): The operator to apply to the last checked in date.

    Returns:
        BaseList[CloudAgent]: A list of CloudAgent objects.
    """

    # Ensure positive page_count:
    if page_count != "all" and page_count < 1:
        raise ValueError("page_count must be 'all' or a positive integer.")

    payload = {"_xml_data": prepare_criteria(**kwargs)}

    results = BaseList()
    pulled = 0

    while True:
        response = call_api(
            auth=auth,
            module="cloud_agent",
            endpoint="list_agents",
            payload=payload,
        )

        parsed = xml_parser(response.text)

        if parsed.get("ServiceResponse")["responseCode"] != "SUCCESS":
            combined_error = f"{parsed.get('ServiceResponse')['responseErrorDetails']['errorMessage']}: {parsed.get('ServiceResponse')['responseErrorDetails']['errorResolution']}"
            raise ValueError(combined_error)

        if parsed.get("ServiceResponse").get("count") == "0":
            break

        data = parsed.get("ServiceResponse").get("data").get("HostAsset")

        if not isinstance(data, list):
            data = [data]

        for agent in data:
            results.append(CloudAgent(**agent))

        pulled += 1
        print(f"Pulled page {pulled}...")

        if page_count != "all" and pulled >= page_count:
            print("Page count reached. Returning...")
            break

        if parsed.get("ServiceResponse").get("hasMoreRecords") != "true":
            break

        else:
            # Grab the last asset ID to use as the pagination ID
            last_id = parsed["ServiceResponse"]["lastId"]
            payload["_xml_data"] = prepare_criteria(
                pagination_id=int(last_id) - 1, **kwargs
            )

    return results


def launch_ods(
    auth: BasicAuth,
    asset_id: str,
    scan: Literal["inv", "vuln", "pc", "udc", "sca", "swca"],
    overrideConfigCpu: bool = False,
) -> str:
    """
    Launch an on-demand scan on a Cloud Agent.

    Args:
        auth (BasicAuth): The authentication object containing the user's credentials.
        asset_id (str): The **Asset ID** of the Cloud Agent to scan.
        scan (Literal['inv', 'vuln', 'pc', 'udc', 'sca', 'swca']): The type of scan to launch.
        overrideConfigCpu (bool): If True, override the CPU configuration. Defaults to False.

    Returns:
        str: The response from the API call.
    """

    # Check valid scan type:
    if scan not in ["inv", "vuln", "pc", "udc", "sca", "swca"]:
        raise ValueError(
            f'Invalid scan type {scan}. Valid types are: {["inv", "vuln", "pc", "udc", "sca", "swca"]}.'
        )

    xml_data = (
        '<?xml version="1.0" encoding="UTF-8" ?> <ServiceRequest></ServiceRequest>'
    )

    payload = {
        "placeholder": asset_id,  # For formatting the URL
        "_xml_data": xml_data,  # call_api will override body with this
    }

    params = {
        "scan": TRANSLATION[scan],
        "overrideConfigCpu": overrideConfigCpu,
    }

    response = call_api(
        auth=auth,
        module="cloud_agent",
        endpoint="launch_ods",
        payload=payload,
        params=params,
    )

    parsed = xml_parser(response.text)

    return (
        parsed.get("ServiceResponse").get("responseCode")
        if response.status_code == 200
        else parsed.get("ServiceResponse")
        .get("responseErrorDetails")
        .get("errorMessage")
        + f": {parsed.get('ServiceResponse').get('responseErrorDetails').get('errorResolution')}"
    )


def bulk_launch_ods(
    auth: BasicAuth,
    scan: Literal["inv", "vuln", "pc", "udc", "sca", "swca"],
    ovverideConfigCpu: bool = False,
    **kwargs,
) -> str:
    """
    Launch on-demand scans in bulk based on asset ID, asset name, or tag name.

    Args:
        auth (BasicAuth): The authentication object containing the user's credentials.
        **kwargs: The filters to apply to the bulk purge.

    ## Kwargs:

        asset_id (Union[list[str], str]): A list of asset IDs to purge. HIGHLY RECOMMENDED.
        name (Union[list[str], str]): A list of asset names to purge.
        tagName (str): A comma-separated string of tag names to purge assets under.

    Returns:
        str: The response from the API call.
    """

    payload = {"_xml_data": prepare_criteria(**kwargs)}

    params = {
        "scan": TRANSLATION[scan],
        "overrideConfigCpu": ovverideConfigCpu,
    }

    response = call_api(
        auth=auth,
        module="cloud_agent",
        endpoint="bulk_launch_ods",
        payload=payload,
        params=params,
    )

    parsed = xml_parser(response.text)

    return (
        parsed.get("ServiceResponse").get("responseCode")
        if response.status_code == 200
        else "ERROR: "
        + xml_parser(response.text)
        .get("ServiceResponse")
        .get("responseErrorDetails")
        .get("errorMessage")
        + f": {parsed.get('ServiceResponse').get('responseErrorDetails').get('errorResolution')}"
    )



"""
Pull details on a single resource by its ID
"""

from typing import Union, Literal

from ..base.call_api import call_api
from ..base.base_list import BaseList
from ..auth.token import BasicAuth
from ..exceptions.Exceptions import *
from .data_classes.AWSResources import *
from .data_classes.resource_mappings import *


def get_resource_details(
    auth: BasicAuth,
    provider: Literal["aws", "azure"],
    resourceType: str,
    resourceId: str,
    page_count: Union[int, "all"] = "all",
    **kwargs,
) -> object:
    """
    Get details of a single resource by its ID

    Params:
        auth (BasicAuth): The authentication object.
        provider (Literal['aws', 'azure']): The cloud provider of the resource.
        resourceType (str): The type of resource to get details for.
        resourceId (str): The ID of the resource to get details for.
        page_count (int): The number of pages to return. If 'all', return all pages. Default is 'all'.

    ## Kwargs:

        pageNo (int): The ordered page to start retrieving resources from, or if page_count is 1, the page to retrieve.
        pageSize (int): The number of resources to get per page.
        filter (str): Filter resources by providing a QQL query.
        sort (Literal['lastSyncedOn:asc', 'lastSyncedOn:desc']): Sort the resources by lastSyncedOn in ascending or descending order.
        updated (str): Filter resources by the last updated date. Format is Qualys QQL, like [2024-01-01 ... 2024-12-31], [2024-01-01 ... now-1m]

    Returns:
        object: The response from the API as an object.
    """

    # Check cloud provider is valid
    provider = provider.lower()
    if provider not in ["aws", "azure"]:
        raise QualysAPIError(
            f"Invalid provider {provider}. Valid providers: 'aws' or 'azure'"
        )

    resourceType = resourceType.upper()

    # Handle common names
    resourceType = resourceType.replace(" ", "_")
    for key, value in COMMON_NAMES[provider].items():
        if resourceType in value:
            resourceType = key
            break

    if resourceType not in VALID_RESOURCETYPES[provider]:
        raise ValueError(
            f"Invalid resource type for provider {provider}. Valid resource types are: {VALID_RESOURCETYPES[provider]}"
        )

    kwargs["placeholder"] = resourceType
    kwargs["resourceid"] = resourceId
    kwargs["cloudprovider"] = provider.upper()

    response = call_api(
        auth=auth,
        module="cloudview",
        endpoint="get_resource_details",
        params=kwargs,
    )

    if response.status_code not in [200, 400, 404]:
        if not response.text:
            raise QualysAPIError(
                "An error occurred while retrieving the resource details. Most likely, a parameter you set is incorrect."
            )
        else:
            raise QualysAPIError(response.json())

    j = response.json()

    if response.status_code == 200:
        resource_class = resource_map.get(resourceType, None)
        if resource_class:
            if "type" in j.keys():
                j["_type"] = j.pop("type")
            obj = resource_class(**j)
        else:
            raise ValueError(
                f"Invalid resource type {resourceType} for provider {provider}. Valid resource types are:\n{VALID_RESOURCETYPES[provider]}"
            )
        return obj

    else:
        j["resourceType"] = resourceType
        raise QualysAPIError(j)



from typing import Any, Dict, List, Literal, Optional, Union
from html import unescape

from xmltodict import unparse


def unparse_to_xml_str(data: Dict[str, Any]) -> str:
    """
    Convert a dictionary to an XML string.

    Args:
         data (Dict[str, Any]): The dictionary to be converted.

    Returns:
         str: The XML string.
    """
    return (
        unescape(unparse(data, full_document=False, pretty=True))
        .replace("True", "true")
        .replace("False", "false")
    )


def remove_none_values(data: Dict[str, Any]) -> Dict[str, Any]:
    if not isinstance(data, dict):
        return data

    # Create a new dictionary to avoid modifying the original while iterating
    new_dict = {}

    for key, value in data.items():
        if isinstance(value, dict):
            # Recursively clean the dictionary
            cleaned_dict = remove_none_values(value)
            if cleaned_dict:  # Only add non-empty dictionaries
                new_dict[key] = cleaned_dict
        elif isinstance(value, list):
            # Recursively clean each item in the list
            cleaned_list = [
                remove_none_values(item) for item in value if item or item == 0
            ]
            if cleaned_list:  # Only add non-empty lists
                new_dict[key] = cleaned_list
        elif (
            value or value == 0 or value is False
        ):  # Consider False and 0 as valid non-null values
            new_dict[key] = value

    return new_dict


def create_service_request(
    filters: Optional[
        List[
            Dict[
                str,
                Union[
                    str,
                    Literal[
                        "CONTAINS",
                        "IN",
                        "EQUALS",
                        "NOT EQUALS",
                        "GREATER",
                        "LESSER",
                        "NONE",
                    ],
                ],
            ]
        ]
    ] = None,
    preferences: Optional[Dict[str, Union[int, float]]] = None,
    data: Optional[Dict[str, Any]] = None,
) -> Dict[str, Any]:
    """
    Creates a dictionary representing a ServiceRequest.

    Args:
        filters (Optional[List[Dict[str, Union[str, Literal["CONTAINS", "IN", "EQUALS", "NOT EQUALS", "GREATER", "LESSER", "NONE"]]]]]): List of filter criteria.
        preferences (Optional[Dict[str, Union[int, float]]]): Preferences for the service request.
        data (Optional[Dict[str, Any]]): Data for the service request.

    Returns:
        Dict[str, Any]: A dictionary matching the ServiceRequest schema.
    """
    xml = {
        "ServiceRequest": {
            "filters": format_filters(filters) if filters else None,
            "preferences": format_preferences(preferences) if preferences else None,
            "data": {
                "WebAppAuthRecord": (
                    format_data(data).pop("WebAppAuthRecord") if data else None
                )
            },
        }
    }

    return remove_none_values(xml)


def format_filters(
    filters: List[
        Dict[
            str,
            Union[
                str,
                Literal[
                    "CONTAINS",
                    "IN",
                    "EQUALS",
                    "NOT EQUALS",
                    "GREATER",
                    "LESSER",
                    "NONE",
                ],
            ],
        ]
    ]
) -> Dict[str, Any]:
    """
    Formats the filters for the ServiceRequest.

    Args:
        filters (List[Dict[str, Union[str, Literal["CONTAINS", "IN", "EQUALS", "NOT EQUALS", "GREATER", "LESSER", "NONE"]]]]): List of filter criteria.

    Returns:
        Dict[str, Any]: A dictionary representing the filters.
    """
    for filter_ in filters:
        if not isinstance(filter_.get("field"), str):
            raise ValueError("Field must be a string.")
        if filter_.get("operator") not in [
            "CONTAINS",
            "IN",
            "EQUALS",
            "NOT EQUALS",
            "GREATER",
            "LESSER",
            "NONE",
        ]:
            raise ValueError("Invalid operator.")
        if not isinstance(filter_.get("value"), str):
            raise ValueError("Value must be a string.")
    return {
        "Criteria": [
            {"field": f["field"], "operator": f["operator"], "value": f["value"]}
            for f in filters
        ]
    }


def format_preferences(
    preferences: Dict[str, Union[int, float]]
) -> Dict[str, Union[int, float]]:
    """
    Formats the preferences for the ServiceRequest.

    Args:
        preferences (Dict[str, Union[int, float]]): Preferences for the service request.

    Returns:
        Dict[str, Union[int, float]]: A dictionary representing the preferences.
    """
    if "startFromId" in preferences and not isinstance(preferences["startFromId"], int):
        raise ValueError("startFromId must be an integer.")
    if "startFromOffset" in preferences and not isinstance(
        preferences["startFromOffset"], int
    ):
        raise ValueError("startFromOffset must be an integer.")
    if "limitResults" in preferences and not isinstance(
        preferences["limitResults"], int
    ):
        raise ValueError("limitResults must be an integer.")
    return preferences


def format_data(data: Dict[str, Any]) -> Dict[str, Any]:
    """
    Formats the data for the ServiceRequest.

    Args:
        data (Dict[str, Any]): Data for the service request.

    Returns:
        Dict[str, Any]: A dictionary representing the data.
    """
    if "WebAppAuthRecord" in data:
        data["WebAppAuthRecord"] = format_web_app_auth_record(data["WebAppAuthRecord"])
    return data


def format_web_app_auth_record(record: Dict[str, Any]) -> Dict[str, Any]:
    """
    Formats the WebAppAuthRecord.

    Args:
        record (Dict[str, Any]): WebAppAuthRecord data.

    Returns:
        Dict[str, Any]: A dictionary representing the WebAppAuthRecord.
    """
    VALID = ["formRecord", "serverRecord", "oauth2Record", "tags", "comments", "name"]
    for key in record.keys():
        if key not in VALID:
            raise ValueError(f"Invalid key: {key}. Must be one of: {VALID}")

    if "formRecord" in record:
        record["formRecord"] = format_web_app_auth_form_record(record["formRecord"])
    if "serverRecord" in record:
        record["serverRecord"] = format_web_app_auth_server_record(
            record["serverRecord"]
        )
    if "oauth2Record" in record:
        record["oauth2Record"] = format_web_app_auth_oauth2_record(
            record["oauth2Record"]
        )
    if "tags" in record:
        record["tags"] = format_tag_list(record["tags"])
    if "comments" in record:
        record["comments"] = format_comment_list(record["comments"])
    return record


def format_web_app_auth_form_record(record: Dict[str, Any]) -> Dict[str, Any]:
    """
    Formats the WebAppAuthFormRecord.

    Args:
        record (Dict[str, Any]): WebAppAuthFormRecord data.

    Returns:
        Dict[str, Any]: A dictionary representing the WebAppAuthFormRecord.
    """
    if "type" in record and record["type"] not in [
        "NONE",
        "STANDARD",
        "CUSTOM",
        "SELENIUM",
    ]:
        raise ValueError(
            f"Invalid type {record['type']} for WebAppAuthFormRecord. Must be one of: {['NONE', 'STANDARD', 'CUSTOM', 'SELENIUM']}"
        )
    if "fields" in record:
        record["fields"] = format_web_app_auth_form_record_field_list(record["fields"])
    return record


def format_web_app_auth_form_record_field_list(
    fields: Dict[str, List[Dict[str, Any]]]
) -> Dict[str, Any]:
    """
    Formats the WebAppAuthFormRecordFieldList.

    Args:
        fields (Dict[str, List[Dict[str, Any]]]): WebAppAuthFormRecordFieldList data.

    Returns:
        Dict[str, Any]: A dictionary representing the WebAppAuthFormRecordFieldList.
    """
    return {
        "set": "".join(
            [
                f'\n<WebAppAuthFormRecordField><name>{field.get("name")}</name>\n<value>{field.get("value")}</value></WebAppAuthFormRecordField>\n'
                for field in fields.get("set", [])
            ]
        ),
        "add": "".join(
            [
                f'\n<WebAppAuthFormRecordField><name>{field.get("name")}</name>\n<secured>{field.get("secured")}</secured>\n<value>{field.get("value")}</value></WebAppAuthFormRecordField>\n'
                for field in fields.get("add", [])
            ]
        ),
        "remove": "".join(
            [
                f'\n<WebAppAuthFormRecordField><name>{field.get("name")}</name>\n<secured>{field.get("secured")}</secured>\n<value>{field.get("value")}</value></WebAppAuthFormRecordField>\n'
                for field in fields.get("remove", [])
            ]
        ),
        "update": "".join(
            [
                f'\n<WebAppAuthFormRecordField><name>{field.get("name")}</name>\n<secured>{field.get("secured")}</secured>\n<value>{field.get("value")}</value></WebAppAuthFormRecordField>\n'
                for field in fields.get("update", [])
            ]
        ),
    }


def format_web_app_auth_form_record_field(field: Dict[str, Any]) -> Dict[str, Any]:
    """
    Formats a WebAppAuthFormRecordField.

    Args:
        field (Dict[str, Any]): WebAppAuthFormRecordField data.

    Returns:
        Dict[str, Any]: A dictionary representing the WebAppAuthFormRecordField.
    """
    if "name" in field and not isinstance(field["name"], str):
        raise ValueError("Field name must be a string.")
    if "secured" in field and not isinstance(field["secured"], bool):
        raise ValueError("Field secured must be a boolean.")
    if "value" in field and not isinstance(field["value"], str):
        raise ValueError("Field value must be a string.")
    # If <SeleniumScript>, we need to validate the data element:
    if "data" in field and not isinstance(field["data"], str):
        try:
            field["data"] = unparse(field["data"])
        except Exception as e:
            raise ValueError(f"Field data must be a string or an XNL-like dictionary.")
    return field


def format_web_app_auth_server_record(record: Dict[str, Any]) -> Dict[str, Any]:
    """
    Formats the WebAppAuthServerRecord.

    Args:
        record (Dict[str, Any]): WebAppAuthServerRecord data.

    Returns:
        Dict[str, Any]: A dictionary representing the WebAppAuthServerRecord.
    """
    if "fields" in record and (
        record["fields"].get("set")
        or record["fields"].get("add")
        or record["fields"].get("remove")
        or record["fields"].get("update")
    ):
        record["fields"] = format_web_app_auth_server_record_field_list(
            record["fields"]
        )
    return record


def format_web_app_auth_server_record_field_list(
    fields: Dict[str, List[Dict[str, Any]]]
) -> Dict[str, Any]:
    """
    Formats the WebAppAuthServerRecordFieldList.

    Args:
        fields (Dict[str, List[Dict[str, Any]]]): WebAppAuthServerRecordFieldList data.

    Returns:
        Dict[str, Any]: A dictionary representing the WebAppAuthServerRecordFieldList.
    """
    return {
        "set": [
            format_web_app_auth_server_record_field(field, "set")
            for field in fields.get("set", [])
        ],
        "add": [
            format_web_app_auth_server_record_field(field, "add")
            for field in fields.get("add", [])
        ],
        "remove": [
            format_web_app_auth_server_record_field(field, "remove")
            for field in fields.get("remove", [])
        ],
        "update": [
            format_web_app_auth_server_record_field(field, "update")
            for field in fields.get("update", [])
        ],
    }


def format_web_app_auth_server_record_field(
    field: Dict[str, Any], command: Literal["set", "remove", "add", "update"]
) -> Dict[str, Any]:
    """
    Formats a WebAppAuthServerRecordField.

    Args:
        field (Dict[str, Any]): WebAppAuthServerRecordField data.
        command (Literal["set", "remove", "add", "update"]): The command to use as an element.

    Returns:
        Dict[str, Any]: A dictionary representing the WebAppAuthServerRecordField.
    """
    if "type" in field and field["type"] not in ["BASIC", "DIGEST", "NTLM"]:
        raise ValueError("Invalid type for WebAppAuthServerRecordField.")
    if "domain" in field and not isinstance(field["domain"], str):
        raise ValueError("Field domain must be a string.")
    if "username" in field and not isinstance(field["username"], str):
        raise ValueError("Field username must be a string.")
    if "password" in field and not isinstance(field["password"], str):
        raise ValueError("Field password must be a string.")

    match command:
        case "set":
            return "".join(
                [
                    f'\n<WebAppAuthServerRecordField>\n<domain>{field.get("domain")}</domain>\n<username>{field.get("username")}</username>\n<password>{field.get("password")}</password></WebAppAuthServerRecordField>\n'
                ]
            )
        case "add":
            return "".join(
                [
                    f'\n<WebAppAuthServerRecordField>\n<domain>{field.get("domain")}</domain>\n<username>{field.get("username")}</username>\n<password>{field.get("password")}</password></WebAppAuthServerRecordField>\n'
                ]
            )
        case "remove":
            return "".join(
                [
                    f'\n<WebAppAuthServerRecordField>\n<domain>{field.get("domain")}</domain>\n<username>{field.get("username")}</username>\n<password>{field.get("password")}</password></WebAppAuthServerRecordField>\n'
                ]
            )
        case "update":
            return "".join(
                [
                    f'\n<WebAppAuthServerRecordField>\n<domain>{field.get("domain")}</domain>\n<username>{field.get("username")}</username>\n<password>{field.get("password")}</password></WebAppAuthServerRecordField>\n'
                ]
            )
        case _:
            raise ValueError("Invalid command for WebAppAuthServerRecordField.")


def format_web_app_auth_oauth2_record(record: Dict[str, Any]) -> Dict[str, Any]:
    """
    Formats the WebAppAuthOAuth2Record.

    Args:
        record (Dict[str, Any]): WebAppAuthOAuth2Record data.

    Returns:
        Dict[str, Any]: A dictionary representing the WebAppAuthOAuth2Record.
    """
    if "grantType" in record and record["grantType"] not in [
        "NONE",
        "AUTH_CODE",
        "IMPLICIT",
        "PASSWORD",
        "CLIENT_CREDS",
    ]:
        raise ValueError("Invalid grantType for WebAppAuthOAuth2Record.")
    if (
        record.get("seleniumScript")
        and record["seleniumScript"].get("data")
        and not isinstance(record["seleniumScript"]["data"], str)
    ):
        try:
            record["seleniumScript"]["data"] = unparse(record["seleniumScript"]["data"])
        except Exception as e:
            raise ValueError(
                f"SeleniumScript data must be a string or an XML-like dictionary."
            )
    return record


def format_tag_list(tags: Dict[str, List[Dict[str, Any]]]) -> Dict[str, Any]:
    """
    Formats the TagList.

    Args:
        tags (Dict[str, List[Dict[str, Any]]]): TagList data.

    Returns:
        Dict[str, Any]: A dictionary representing the TagList.
    """
    """return {
        "set": tags.get("set", []),
        "add": tags.get("add", []),
        "remove": tags.get("remove", []),
    }"""

    # BUG: Above needs to be refactored to build a "".join() string of the XML elements for the tags.
    # due to xmltodict's behavior.
    # CORRECTION:
    return {
        "set": "".join(
            [f'\n<Tag>\n<id>{tag["id"]}</id>\n</Tag>' for tag in tags.get("set", [])]
        ),
        "add": "".join(
            [f'\n<Tag>\n<id>{tag["id"]}</id>\n</Tag>' for tag in tags.get("add", [])]
        ),
        "remove": "".join(
            [f'\n<Tag>\n<id>{tag["id"]}</id>\n</Tag>' for tag in tags.get("remove", [])]
        ),
        "update": "".join(
            [f'\n<Tag>\n<id>{tag["id"]}</id>\n</Tag>' for tag in tags.get("update", [])]
        ),
    }


def format_comment_list(comments: Dict[str, List[str]]) -> Dict[str, Any]:
    """
    Formats the CommentList.

    Args:
        comments (Dict[str, List[str]]): CommentList data.

    Returns:
        Dict[str, Any]: A dictionary representing the CommentList.
    """
    # Strange xmltodict behavior forces us to build a string of the XML elements for the comments.
    return {
        "set": "".join(
            [
                f"\n<Comment>\n<contents>{comment}</contents>\n</Comment>"
                for comment in comments.get("set", [])
            ]
        ),
        "add": "".join(
            [
                f"\n<Comment>\n<contents>{comment}</contents>\n</Comment>"
                for comment in comments.get("add", [])
            ]
        ),
        "remove": "".join(
            [
                f"\n<Comment>\n<contents>{comment}</contents>\n</Comment>"
                for comment in comments.get("remove", [])
            ]
        ),
        "update": "".join(
            [
                f"\n<Comment>\n<contents>{comment}</contents>\n</Comment>"
                for comment in comments.get("update", [])
            ]
        ),
    }



"""
activity_log.py - Contains the user-facing functionality for interacting with the Qualys VMDR Activity Log API.
"""

from urllib.parse import parse_qs
from re import compile, DOTALL
from csv import DictReader
from typing import Union

from ..base.call_api import call_api
from ..auth.basic import BasicAuth
from ..base.base_list import BaseList
from ..vmdr.data_classes.activity_log import ActivityLog


def extract_sections(csv_data: str) -> tuple[Union[str, None], Union[str, None]]:
    """
    Split the CSV data into the body and footer sections.

    Args:
        csv_data (str): The raw requests.response.text data.

    Returns:
        tuple[Union[str, None], Union[str, None]]: The body and footer sections, in that order. None if not found.
    """

    body_pattern = compile(
        r"----BEGIN_RESPONSE_BODY_CSV\n(.*?)\n----END_RESPONSE_BODY_CSV", DOTALL
    )
    footer_patern = compile(
        r"----BEGIN_RESPONSE_FOOTER_CSV\n(.*?)\n----END_RESPONSE_FOOTER_CSV", DOTALL
    )

    body_match = body_pattern.search(csv_data)
    body_section = body_match.group(1) if body_match else None

    footer_match = footer_patern.search(csv_data)
    footer_section = footer_match.group(1) if footer_match else None

    return body_section, footer_section


def get_activity_log(
    auth: BasicAuth, page_count: Union[int, "all"] = "all", **kwargs
) -> BaseList[ActivityLog]:
    """
    Get the activity log for the subscription.

    Args:
        auth (BasicAuth): The BasicAuth object containing the user's credentials.
        page_count (Union[int, 'all']): The number of pages to pull. Defaults to 'all'.
        **kwargs: Additional parameters to pass to the API.

    :Kwargs:
        user_action (str): Filter by user action.
        action_details (str): Filter by action details.
        username (str): Filter by username.
        since_datetime (str): Filter by date and time since. Formatted like ```YYYY-MM-DD HH:ii:ss```.
        until_datetime (str): Filter by date and time until. Formatted like ```YYYY-MM-DD HH:ii:ss```.
        user_role (str): Filter by user role.
        truncation_limit (int): The maximum number of characters to return in the details field.

    Returns:
        BaseList[ActivityLog]: The list of activity log entries.
    """

    responses = BaseList()
    pulled = 0
    params = {"action": "list", "output_format": "csv"}

    if kwargs:
        params.update(kwargs)

    while True:
        # make the request:
        response = call_api(
            auth=auth,
            module="vmdr",
            endpoint="get_activity_log",
            params=params,
            headers={"X-Requested-With": "qualysdk SDK"},
        )
        if response.status_code != 200:
            print("No data returned.")
            return responses

        # Rip the data out of the header/footer/warning comments:
        data, pagination_data = extract_sections(response.text)

        if not data:
            print("No data returned.")
            return responses

        data = DictReader(data.splitlines())

        for row in data:
            # Keys with a space in the name need to be renamed to
            # have underscores instead of spaces.
            row = {k.replace(" ", "_"): v for k, v in row.items()}
            responses.append(ActivityLog.from_dict(row))

        # Check for pagination:
        if pagination_data:
            url = pagination_data.split(",")[~0]

            # Parse the params out of the URL:
            url_params = parse_qs(url)

            # Look for the id_max parameter and update the params:
            if "id_max" in url_params:
                params["id_max"] = url_params["id_max"][0].strip().replace('"', "")
                print(
                    f"Pagination detected. Pulling next page with id_max: {params['id_max']}"
                )
            else:
                print("No more pages to pull.")
                break

        pulled += 1

        if page_count != "all" and pulled >= page_count:
            print("Page count reached.")
            break

    return responses



"""
Contains the code to interact with Docker hosts.
"""

from typing import Union
from urllib.parse import parse_qs

from ..data_classes.container import Container
from ...auth.token import TokenAuth
from ...base.call_api import call_api
from ...base.base_list import BaseList
from ...exceptions.Exceptions import *


def list_containers(
    auth: TokenAuth, page_count: Union[int, "all"] = "all", **kwargs
) -> BaseList[Container]:
    """
    Get a list of all containers according to kwargs.

    Args:
        auth (TokenAuth): The authentication token.
        page_count (Union[int, 'all'] = 'all'): How many pages of results to retrieve.
        **kwargs: Any additional arguments to pass to the API.

    ## Kwargs:

    - filter (str): A filter to apply to the list using Qualys Container Security QQL.
    - paginationQuery (str): The pagination query to use. NOTE: The SDK handles this for you.
    - limit (int): The maximum number of results to return per page.

    Returns:
        BaseList[Container]: A list of Container objects.
    """

    # Check if page_count is valid:
    if page_count != "all" and not isinstance(page_count, int):
        raise ValueError("page_count must be an integer or 'all'.")

    results = BaseList()
    pages_pulled = 0

    while True:
        # Pull the data:
        response = call_api(auth, "containersecurity", "list_containers", params=kwargs)

        if response.status_code != 200 and response.text:
            raise QualysAPIError(response.json())
        elif response.status_code != 200 and not response.text:
            raise QualysAPIError(
                f"An error occurred with status code {response.status_code}. Qualys did not return any additional information."
            )

        data = response.json()

        # Check if the data is empty:
        if not data.get("data"):
            break

        if isinstance(data.get("data"), dict):
            # If the data is a dict, convert it to a list of dicts:
            data["data"] = [data["data"]]

        # Add the data to the results:
        for container in data["data"]:
            results.append(Container(**container))

        pages_pulled += 1

        # Check if we need to pull more pages:
        if page_count != "all" and pages_pulled >= page_count:
            print(f"Page count reached. Returning {pages_pulled} pages of containers.")
            break

        # Check the response headers for the next page:
        headers = dict(response.headers)
        if not headers.get("Link"):
            break
        else:
            # Use parse_qs to get the pagination query:
            pagination_query = parse_qs(headers["Link"].split(";")[0].strip("<>"))[
                "paginationQuery"
            ][0]
            kwargs["paginationQuery"] = pagination_query

    return results


def get_container_details(auth: TokenAuth, containerSha: str) -> Container:
    """
    Get details of a single container instance.

    Args:
        auth (TokenAuth): The authentication token.
        containerSha (str): The SHA hash of the container.

    Returns:
        Container: Container object with all attributes populated.
    """

    response = call_api(
        auth,
        module="containersecurity",
        endpoint="get_container_details",
        params={"placeholder": containerSha},
    )

    # Check for valid response:
    if response.status_code != 200:
        raise QualysAPIError(response.json())

    return Container.from_dict(response.json())



"""
bugtraq.py - contains the BugTraq dataclass for the Qualys VMDR module.
"""

from dataclasses import dataclass, field, asdict
from typing import *


@dataclass
class Bugtraq:
    """
    BugTraq - represents a single BugTraq entry in a BugTraqList.
    """

    ID: int = field(metadata={"description": "The BugTraq ID."})
    URL: str = field(
        metadata={"description": "The URL of the BugTraq."}, default="", compare=False
    )

    def __post_init__(self):
        # make sure that the ID is an integer:
        if not isinstance(self.ID, int):
            self.ID = int(self.ID)
        # and that url is a string:
        if not isinstance(self.URL, str):
            self.URL = str(self.URL)

    def __str__(self):
        return str(self.ID)

    def __dict__(self):
        return asdict(self)

    def __contains__(self, item):
        return item in self.ID or item in self.URL

    def copy(self):
        return Bugtraq(ID=self.ID, URL=self.URL)

    def is_id(self, id: int):
        return self.ID == id

    def is_url(self, url: str):
        return self.URL == url

    def to_dict(self):
        return asdict(self)

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @classmethod
    def from_dict(cls, data: dict):
        """
        from_dict - create a BugTraq object from a dictionary.

        This function is used to create a BugTraq object from a dictionary.
        """
        required_keys = {"ID"}
        if not required_keys.issubset(data.keys()):
            raise ValueError(
                f"Dictionary must contain the following keys: {required_keys}"
            )

        return cls(**data)



"""
hosts.py - contains the VMDRHosts class for the qualysdk package.
"""

from dataclasses import dataclass, field, asdict
from typing import *
from datetime import datetime

from ipaddress import IPv4Address, IPv6Address

from ...base.base_list import BaseList
from .tag import Tag, CloudTag
from .detection import Detection


@dataclass(order=True)
class VMDRHost:
    """
    Host - represents a Qualys VMDR host record.
    """

    ID: Union[str, int] = field(
        default=None, metadata={"description": "The HOST ID of the host."}
    )  # this is the host ID, not the asset ID. add to post init to cast to int
    ASSET_ID: Union[str, int] = field(
        default=None, metadata={"description": "The asset ID of the host."}
    )  # add to post init to cast to int
    IP: Union[str, IPv4Address] = field(
        default=None,
        metadata={"description": "The IP address of the host."},
        compare=False,
    )
    IPV6: Union[str, IPv6Address] = field(
        default=None,
        metadata={"description": "The IPv6 address of the host."},
        compare=False,
    )
    TRACKING_METHOD: str = field(
        default=None,
        metadata={"description": "The tracking method of the host."},
        compare=False,
    )
    DNS: str = field(
        default=None,
        metadata={"description": "The DNS name of the host."},
        compare=False,
    )

    # DNS_DATA is a dictionary containing the keys HOSTNAME, DOMAIN, and FQDN. We will make these keys attributes of the class.
    DNS_DATA: Dict[str, str] = field(
        default=None,
        metadata={"description": "The DNS data of the host."},
        compare=False,
    )

    NETBIOS: str = field(
        default=None,
        metadata={"description": "The NetBIOS name of the host."},
        compare=False,
    )
    OS: str = field(
        default=None,
        metadata={"description": "The operating system of the host."},
        compare=False,
    )
    QG_HOSTID: str = field(
        default=None, metadata={"description": "The QualysGuard host ID of the host."}
    )
    LAST_BOOT: Union[str, datetime] = field(
        default=None,
        metadata={"description": "The last boot time of the host."},
        compare=False,
    )  # add to post init to cast to datetime

    LAST_SCAN_DATETIME: Union[str, datetime] = field(
        default=None,
        metadata={"description": "The last scan date of the host."},
        compare=False,
    )  # add to post init to cast to datetime

    SERIAL_NUMBER: str = field(
        default=None,
        metadata={"description": "The serial number of the host."},
        compare=False,
    )
    HARDWARE_UUID: str = field(
        default=None, metadata={"description": "The hardware UUID of the host."}
    )
    FIRST_FOUND_DATE: Union[str, datetime] = field(
        default=None,
        metadata={"description": "The first found date of the host."},
        compare=False,
    )  # add to post init to cast to datetime
    LAST_ACTIVITY: Union[str, datetime] = field(
        default=None,
        metadata={"description": "The last activity date of the host."},
        compare=False,
    )
    LAST_ACTIVITY_DATE: Union[str, datetime] = field(
        default=None,
        metadata={"description": "The last activity date of the host."},
        compare=False,
    )  # add to post init to cast to datetime
    AGENT_STATUS: str = field(
        default=None,
        metadata={"description": "The agent status of the host."},
        compare=False,
    )
    CLOUD_AGENT_RUNNING_ON: str = field(
        default=None,
        metadata={"description": "The cloud agent running on of the host."},
        compare=False,
    )
    TAGS: Union[dict, BaseList] = field(
        default=None, metadata={"description": "The tags of the host."}, compare=False
    )  # add to post init to convert to BaseList (look at GAV lists for how to do this), as well as make a Tag class
    LAST_VULN_SCAN_DATETIME: Union[str, datetime] = field(
        default=None,
        metadata={"description": "The last vulnerability scan date of the host."},
        compare=False,
    )  # add to post init to cast to datetime
    LAST_VULN_SCAN_DATE: Union[str, datetime] = field(
        default=None,
        metadata={"description": "The last vulnerability scan date of the host."},
        compare=False,
    )  # add to post init to cast to datetime
    LAST_VM_SCANNED_DATE: Union[str, datetime] = field(
        default=None,
        metadata={"description": "The last VM scan date of the host."},
        compare=False,
    )
    LAST_VM_AUTH_SCANNED_DURATION: Union[str, int] = field(
        default=None,
        metadata={"description": "The last VM scanned duration of the host."},
        compare=False,
    )  # add to post init to cast to int
    LAST_VM_SCANNED_DURATION: Union[str, int] = field(
        default=None,
        metadata={"description": "The last VM scanned duration of the host."},
        compare=False,
    )  # add to post init to cast to int
    LAST_VM_AUTH_SCANNED_DATE: Union[str, datetime] = field(
        default=None,
        metadata={"description": "The last VM auth scanned date of the host."},
        compare=False,
    )  # add to post init to cast to datetime
    LAST_COMPLIANCE_SCAN_DATETIME: Union[str, datetime] = field(
        default=None,
        metadata={"description": "The last compliance scan date of the host."},
        compare=False,
    )  # add to post init to cast to datetime

    LAST_PC_SCANNED_DATE: Union[str, datetime] = field(
        default=None,
        metadata={"description": "The last PC scanned date of the host."},
        compare=False,
    )  # add to post init to cast to datetime

    ASSET_GROUP_IDS: List[str] = field(
        default=None,
        metadata={"description": "The asset group IDs of the host."},
        compare=False,
    )
    USER_DEF: dict = field(
        default=None,
        metadata={"description": "The user def of the host."},
        compare=False,
    )
    OWNER: str = field(
        default=None, metadata={"description": "The owner of the host."}, compare=False
    )

    # CLOUD HOST FIELDS:
    CLOUD_PROVIDER: str = field(
        default=None,
        metadata={"description": "The cloud provider of the host."},
        compare=False,
    )
    CLOUD_SERVICE: str = field(
        default=None,
        metadata={"description": "The cloud service of the host."},
        compare=False,
    )
    CLOUD_RESOURCE_ID: str = field(
        default=None,
        metadata={"description": "The cloud resource ID of the host."},
        compare=False,
    )
    EC2_INSTANCE_ID: str = field(
        default=None,
        metadata={"description": "The EC2 instance ID of the host."},
        compare=False,
    )
    CLOUD_PROVIDER_TAGS: Union[dict, BaseList] = field(
        default=None,
        metadata={"description": "The cloud provider tags of the host."},
        compare=False,
    )

    # this METADATA field contains a lot of nested data, so we will need to parse it out in post init. The immediate key underneath is the cloud service, i.e. EC2, AZURE, etc.
    # and gets parsed in post init.
    METADATA: dict = field(
        default=None,
        metadata={"description": "The metadata of the host."},
        compare=False,
    )

    DETECTION_LIST: Optional[BaseList[Detection]] = field(
        default=None,
        metadata={"description": "The detection list of the host."},
        compare=False,
    )

    ASSET_RISK_SCORE: Optional[Union[str, int]] = field(
        default=None,
        metadata={"description": "The asset risk score of the host."},
        compare=False,
    )

    TRURISK_SCORE: Optional[Union[str, int]] = field(
        default=None,
        metadata={"description": "The TruRisk score of the host."},
        compare=False,
    )

    TRURISK_SCORE_FACTORS: Optional[dict] = field(
        default=None,
        metadata={"description": "The TruRisk score factors of the host."},
        compare=False,
    )

    ASSET_CRITICALITY_SCORE: Optional[Union[str, int]] = field(
        default=None,
        metadata={"description": "The asset criticality score of the host."},
        compare=False,
    )

    CLOUD_GROUP_NAME: Optional[str] = field(
        default=None,
        metadata={"description": "The cloud group name of the host."},
        compare=False,
    )

    CLOUD_INSTANCE_STATE: Optional[str] = field(
        default=None,
        metadata={"description": "The cloud instance state of the host."},
        compare=False,
    )

    CLOUD_INSTANCE_TYPE: Optional[str] = field(
        default=None,
        metadata={"description": "The cloud instance type of the host."},
        compare=False,
    )

    CLOUD_IS_SPOT_INSTANCE: Optional[bool] = field(
        default=None,
        metadata={"description": "The cloud is spot instance of the host."},
        compare=False,
    )

    CLOUD_ARCHITECTURE: Optional[str] = field(
        default=None,
        metadata={"description": "The cloud architecture of the host."},
        compare=False,
    )

    CLOUD_IMAGE_ID: Optional[str] = field(
        default=None,
        metadata={"description": "The cloud image ID of the host."},
        compare=False,
    )

    CLOUD_REGION: Optional[str] = field(
        default=None,
        metadata={"description": "The cloud region of the host."},
        compare=False,
    )

    CLOUD_AMI_ID: Optional[str] = field(
        default=None,
        metadata={"description": "The cloud AMI ID of the host."},
        compare=False,
    )

    CLOUD_PUBLIC_HOSTNAME: Optional[str] = field(
        default=None,
        metadata={"description": "The cloud public hostname of the host."},
        compare=False,
    )

    CLOUD_PUBLIC_IPV4: Optional[Union[str, IPv4Address]] = field(
        default=None,
        metadata={"description": "The cloud public IPv4 of the host."},
        compare=False,
    )

    CLOUD_ACCOUNT_ID: Optional[Union[str, int]] = field(
        default=None,
        metadata={"description": "The cloud account ID of the host."},
        compare=False,
    )

    def __post_init__(self):
        """
        Pull up nested dict values as attributes, convert IPs,
        put tags in a BaseList and convert strings to datetime objects.
        """
        DNS_DATA_FIELDS = ["HOSTNAME", "DOMAIN", "FQDN"]
        DATETIME_FIELDS = [
            "LAST_BOOT",
            "FIRST_FOUND_DATE",
            "LAST_ACTIVITY_DATE",
            "LAST_SCAN_DATETIME",
            "LAST_VULN_SCAN_DATETIME",
            "LAST_VM_SCANNED_DATE",
            "LAST_VM_AUTH_SCANNED_DATE",
            "LAST_VM_AUTH_SCANNED_DATE",
            "LAST_COMPLIANCE_SCAN_DATETIME",
            "LAST_VULN_SCAN_DATE",
            "LAST_ACTIVITY",
            "LAST_PC_SCANNED_DATE",
        ]
        INT_FIELDS = [
            "ID",
            "ASSET_ID",
            "LAST_VM_SCANNED_DURATION",
            "ASSET_RISK_SCORE",
            "TRURISK_SCORE",
            "ASSET_CRITICALITY_SCORE",
        ]

        if self.DNS_DATA:
            for field in DNS_DATA_FIELDS:
                setattr(self, field, self.DNS_DATA.get(field))

        if self.IP and not isinstance(self.IP, IPv4Address):
            self.IP = IPv4Address(self.IP)

        if self.IPV6 and not isinstance(self.IPV6, IPv6Address):
            self.IPV6 = IPv6Address(self.IPV6)

        for DATE_FIELD in DATETIME_FIELDS:
            if getattr(self, DATE_FIELD) and not isinstance(
                getattr(self, DATE_FIELD), datetime
            ):
                setattr(
                    self, DATE_FIELD, datetime.fromisoformat(getattr(self, DATE_FIELD))
                )

        if self.TRURISK_SCORE_FACTORS:
            s = ""
            for sev_level in self.TRURISK_SCORE_FACTORS.get("VULN_COUNT"):
                s += f"sev_{sev_level.get('@qds_severity')}: {sev_level.get('#text')}, "
            # Pinch off the trailing comma and space:
            s = s[:-2]
            self.TRURISK_SCORE_FACTORS = s

        if self.TAGS:
            # if 'TAG' key's value is a list, it is a list of tag dicts. if it is a single tag dict, it is just a single tag.
            if isinstance(self.TAGS["TAG"], list):
                self.TAGS = BaseList([Tag.from_dict(tag) for tag in self.TAGS["TAG"]])
            else:  # if it is a single tag dict:
                self.TAGS = BaseList([Tag.from_dict(self.TAGS["TAG"])])

        if self.CLOUD_PROVIDER_TAGS:
            # if 'CLOUD_TAG' key's value is a list, it is a list of tag dicts. if it is a single tag dict, it is just a single tag.
            if isinstance(self.CLOUD_PROVIDER_TAGS["CLOUD_TAG"], list):
                self.CLOUD_PROVIDER_TAGS = BaseList(
                    [
                        CloudTag.from_dict(tag)
                        for tag in self.CLOUD_PROVIDER_TAGS["CLOUD_TAG"]
                    ]
                )
            else:  # if it is a single tag dict:
                self.CLOUD_PROVIDER_TAGS = BaseList(
                    [CloudTag.from_dict(self.CLOUD_PROVIDER_TAGS["CLOUD_TAG"])]
                )

        # CLOUD SPECIFIC FIELDS:
        if self.METADATA:
            match self.CLOUD_PROVIDER:
                case "AWS":
                    key_selector = "EC2"
                case "Azure":
                    key_selector = "AZURE"
                case "GCP":
                    key_selector = "GCP"
                case _:
                    raise ValueError(
                        f"Cloud provider {self.CLOUD_PROVIDER} is not supported."
                    )

            # for each tuple, [0] is the dataclass attribute name, [1] is how it is represented in the metadata dict.
            VALID_EC2_KEYS = [
                ("GROUP_NAME", "groupName"),
                ("INSTANCE_STATE", "instanceState"),
                ("INSTANCE_TYPE", "latest/meta-data/instance-type"),
                ("IS_SPOT_INSTANCE", "isSpotInstance"),
                (
                    "ARCHITECTURE",
                    "latest/dynamic/instance-identity/document/architecture",
                ),
                ("IMAGE_ID", "latest/dynamic/instance-identity/document/imageId"),
                ("REGION", "latest/dynamic/instance-identity/document/region"),
                ("AMI_ID", "latest/meta-data/ami-id"),
                ("PUBLIC_HOSTNAME", "latest/meta-data/public-hostname"),
                ("PUBLIC_IPV4", "latest/meta-data/public-ipv4"),
                ("ACCOUNT_ID", "latest/dynamic/instance-identity/document/accountId"),
            ]
            VALID_AZURE_KEYS = [
                ("PUBLIC_IPV4", "latest/meta-data/public-ipv4"),
                ("INSTANCE_STATE", "state"),
                ("GROUP_NAME", "resourceGroupName"),
                ("INSTANCE_TYPE", "vmSize"),
                ("REGION", "location"),
                ("ACCOUNT_ID", "subscriptionId"),
            ]
            VALID_GCP_KEYS = ...
            # map key_selector to the valid keys list:
            VALID_KEYS = {
                "EC2": VALID_EC2_KEYS,
                "AZURE": VALID_AZURE_KEYS,
                # "GCP": VALID_GCP_KEYS #not implemented as i have no access to a GCP environment.
            }

            for key in VALID_KEYS[key_selector]:
                # check for if self.METADATA[key_selector]['ATTRIBUTE'] is a list of dicts. if not, it is just a single dict.
                if isinstance(self.METADATA[key_selector]["ATTRIBUTE"], list):
                    for item in self.METADATA[key_selector]["ATTRIBUTE"]:
                        if item["NAME"] == key[1]:
                            setattr(
                                self,
                                f"CLOUD_{key[0]}",
                                (
                                    item["VALUE"]
                                    if item["VALUE"] not in ["", {}, []]
                                    else None
                                ),
                            )  # if item['VALUE'] seems to leave behind empties, hence the list
                            break
                else:
                    if (
                        self.METADATA[key_selector]["ATTRIBUTE"]["NAME"]
                        and self.METADATA[key_selector]["ATTRIBUTE"]["NAME"] == key
                    ):
                        setattr(
                            self,
                            f"CLOUD_{key[0]}",
                            self.METADATA[key_selector]["ATTRIBUTE"]["VALUE"],
                        )

        for INT_FIELD in INT_FIELDS:
            if getattr(self, INT_FIELD) and not isinstance(
                getattr(self, INT_FIELD), int
            ):
                setattr(self, INT_FIELD, int(getattr(self, INT_FIELD)))

        if self.CLOUD_PUBLIC_IPV4 and not isinstance(
            self.CLOUD_PUBLIC_IPV4, IPv4Address
        ):
            self.CLOUD_PUBLIC_IPV4 = IPv4Address(self.CLOUD_PUBLIC_IPV4)

        if self.CLOUD_IS_SPOT_INSTANCE:
            self.CLOUD_IS_SPOT_INSTANCE = bool(self.CLOUD_IS_SPOT_INSTANCE)

        # check for a detections list and convert it to a BaseList of Detection objects (used in hld):
        if self.DETECTION_LIST:
            detections_bl = BaseList()
            data = self.DETECTION_LIST["DETECTION"]

            if isinstance(data, dict):
                data = [data]

            for detection in data:
                # Append the host's ID attr to the detection dictionary
                # to allow for a relationship:
                detection["ID"] = self.ID
                detections_bl.append(Detection.from_dict(detection))

            self.DETECTION_LIST = detections_bl

    def __str__(self) -> str:
        """
        String representation of the host object.
        """
        return str(self.ID)

    def __int__(self) -> int:
        return self.ID

    def __repr__(self) -> str:
        """
        Breaking the unwritten rule of having repr be something
        that can create the object for terminal space's sake.
        """
        if self.ASSET_ID:
            return f"Host({self.ASSET_ID})"
        elif self.ID:
            return f"Host({self.ID})"
        else:
            # fall back to QG_HostID:
            return f"Host({self.QG_HOSTID})"

    def __iter__(self):
        """
        Allows for iteration over the host object.
        """
        for key, value in self.to_dict().items():
            yield key, value

    def has_agent(self) -> bool:
        """
        Check if the host has an agent.
        """
        return self.QG_HOSTID is not None

    def is_cloud_host(self) -> bool:
        """
        Check if the host is a cloud host.
        """
        return self.CLOUD_PROVIDER is not None

    def is_aws(self) -> bool:
        """
        Check if the host is an AWS host.
        """
        return self.CLOUD_PROVIDER == "AWS"

    def is_azure(self) -> bool:
        """
        Check if the host is an Azure host.
        """
        return self.CLOUD_PROVIDER == "Azure"

    def copy(self) -> "VMDRHost":
        """
        Create a copy of the host object.
        """
        return VMDRHost.from_dict(self.to_dict())

    def valid_values(self) -> list:
        """
        Return a list of keys that have values.
        """
        return {k: v for k, v in self.items() if v}

    def to_dict(self) -> dict:
        """
        Convert the host object to a dictionary.
        """
        return asdict(self)

    def __dict__(self) -> dict:
        """
        Convert the host object to a dictionary.
        """
        return asdict(self)

    def keys(self) -> list:
        """
        Return the keys of the host object.
        """
        return self.to_dict().keys()

    def values(self) -> list:
        """
        Return the values of the host object.
        """
        return self.to_dict().values()

    def items(self) -> list:
        """
        Return the items of the host object.
        """
        return self.to_dict().items()

    @classmethod
    def from_dict(cls, data: dict) -> "VMDRHost":
        """
        Create a VMDRHost object from a dictionary.
        """
        return cls(**data)


@dataclass
class VMDRID:
    """
    ID - represents a Qualys GAV ID record.

    This class is used to represent a Qualys GAV ID record, either a host ID or an asset ID.
    This class is only ever used if details=None in a get_host_list or get_host_list_detections API call.
    """

    ID: Union[str, int] = field(
        metadata={"description": "The asset ID of the host."}
    )  # add to post init to cast to int
    TYPE: Literal["asset", "host"] = field(
        metadata={"description": "The type of ID. Valid values are 'asset' or 'host'."},
        compare=False,
    )  # field is required!

    def __post_init__(self):
        """
        Cast the asset ID and host ID to integers if they are not None.
        """

        if not self.ID:
            raise ValueError("ID attribute cannot be None.")

        self.ID = int(self.ID)

    def __str__(self) -> str:
        return str(self.ID)

    def __int__(self) -> int:
        return self.ID

    def __repr__(self) -> str:
        if self.TYPE == "asset":
            return f"VMDRID({self.ID}, type='asset')"
        else:
            return f"VMDRID({self.ID}, type='host')"

    def __dict__(self) -> dict:
        return asdict(self)

    def to_dict(self) -> dict:
        return asdict(self)

    def __iter__(self):
        """
        Iterate over the fields of the host object.
        """
        for key, value in self.to_dict().items():
            yield key, value

    def values(self):
        """
        Return the values of the object.
        """
        return self.to_dict().values()

    def keys(self):
        """
        Return the keys of the object.
        """
        return self.to_dict().keys()

    @classmethod
    def from_dict(cls, data: dict) -> "VMDRID":
        """
        Create a VMDRID object from a dictionary.
        """
        required_keys = {"ID", "TYPE"}
        if not required_keys.issubset(data.keys()):
            raise ValueError(
                f"Dictionary must contain the following keys: {required_keys}"
            )

        return cls(**data)
